#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
'''

'''

import logging
log = logging.getLogger(__name__)

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Standard library imports
from importlib import import_module
import sys

# External imports

# Intake imports
from intake import __version__
from intake.cli.util import Subcommand

#-----------------------------------------------------------------------------
# API
#-----------------------------------------------------------------------------

class Info(Subcommand):
    ''' Display runtime information related to Intake

    '''

    name = "info"

    def initialize(self):
        pass

    def invoke(self, args):
        print("Python version      :  %s" % sys.version.split('\n')[0])
        print("IPython version     :  %s" % _version_from_module('IPython'))
        print("Tornado version     :  %s" % _version_from_module('tornado', 'version'))
        print("Dask version        :  %s" % _version_from_module('dask'))
        print("Pandas version      :  %s" % _version_from_module('pandas'))
        print("Numpy version       :  %s" % _version_from_module('numpy'))
        print("Intake version      :  %s" % __version__)

def _version_from_module(modname, version_attr="__version__"):
    try:
        mod = import_module(modname)
        return getattr(mod, version_attr)
    except ImportError:
        return "(not installed)"



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import argparse
import logging
import signal
import sys

import tornado.ioloop
import tornado.web

from .server import IntakeServer
logger = logging.getLogger('intake')


def call_exit_on_sigterm(signal, frame):
    sys.exit(0)


def main(argv=None):
    from intake.config import conf
    from intake import open_catalog

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description='Intake Catalog Server')
    parser.add_argument('-p', '--port', type=int, default=conf['port'],
                        help='port number for server to listen on')
    parser.add_argument('--list-entries', action='store_true',
                        help='list catalog entries at startup')
    parser.add_argument('--sys-exit-on-sigterm', action='store_true',
                        help='internal flag used during unit testing to ensure '
                             '.coverage file is written')
    parser.add_argument('catalog_args', metavar='FILE', type=str, nargs='+',
                        help='Name of catalog YAML file')
    parser.add_argument('--flatten', dest='flatten', action='store_true')
    parser.add_argument('--no-flatten', dest='flatten', action='store_false')
    parser.add_argument('--ttl', dest='ttl',type=int, default=60)
    parser.add_argument('-a', '--address', type=str,
                        default=conf.get('address', 'localhost'),
                        help='address to use as a host, defaults to the address '
                        'in the configuration file, if provided otherwise localhost')
    parser.set_defaults(flatten=True)
    args = parser.parse_args(argv[1:])

    if args.sys_exit_on_sigterm:
        signal.signal(signal.SIGTERM, call_exit_on_sigterm)

    logger.info('Creating catalog from:')
    for arg in args.catalog_args:
        logger.info('  - %s' % arg)

    catargs = args.catalog_args
    ttl = args.ttl

    if len(catargs) == 1:
        catalog = open_catalog(catargs[0], ttl=ttl)
        logger.info("catalog_args: %s" % catargs[0])
    else:
        catalog = open_catalog(catargs, flatten=args.flatten, ttl=ttl)
        logger.info("catalog_args: %s" % catargs)
    if args.list_entries:
        # This is not a good idea if the Catalog is huge.
        logger.info('Entries:' + ','.join(list(catalog)))

    logger.info('Listening on %s:%d' % (args.address, args.port))

    server = IntakeServer(catalog)
    app = server.make_app()
    server.start_periodic_functions(close_idle_after=3600.0)

    app.listen(args.port, address=args.address)
    try:
        tornado.ioloop.IOLoop.current().start()
    except KeyboardInterrupt:
        logger.critical("Exiting")
    except Exception as e:
        logger.critical("Exiting due to %s" % e)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



import contextlib
import json
from itertools import islice

from intake.source.base import DataSource


class JSONFileSource(DataSource):
    """
    Read JSON files as a single dictionary or list

    The files can be local or remote. Extra parameters for encoding, etc.,
    go into ``storage_options``.
    """

    name = "json"
    version = "0.0.1"
    container = "python"

    def __init__(
        self,
        urlpath: str,
        text_mode: bool = True,
        text_encoding: str = "utf8",
        compression: str = None,
        read: bool = True,
        metadata: dict = None,
        storage_options: dict = None,
    ):
        """
        Parameters
        ----------
        urlpath : str
            Target file. Can include protocol specified (e.g., "s3://").
        text_mode : bool
            Whether to open the file in text mode, recoding binary
            characters on the fly
        text_encoding : str
            If text_mode is True, apply this encoding. UTF* is by far the most
            common
        compression : str or None
            If given, decompress the file with the given codec on load. Can
            be something like "zip", "gzip", "bz2", or to try to guess from the
            filename, 'infer'
        storage_options: dict
            Options to pass to the file reader backend, including text-specific
            encoding arguments, and parameters specific to the remote
            file-system driver, if using.
        """
        from fsspec.utils import compressions

        VALID_COMPRESSIONS = list(compressions.values()) + ["infer"]

        self._urlpath = urlpath
        self._storage_options = storage_options or {}
        self._dataframe = None
        self._file = None
        self.compression = compression
        if compression is not None:
            if compression not in VALID_COMPRESSIONS:
                raise ValueError(
                    f"Compression value {compression} must be one of {VALID_COMPRESSIONS}"
                )
        self.mode = "rt" if text_mode else "rb"
        self.encoding = text_encoding
        self._read = read

        super(JSONFileSource, self).__init__(metadata=metadata)

    def read(self):
        import fsspec

        urlpath = self._get_cache(self._urlpath)[0]
        with fsspec.open(
            urlpath,
            mode=self.mode,
            encoding=self.encoding,
            compression=self.compression,
            **self._storage_options,
        ) as f:
            return json.load(f)

    def _load_metadata(self):
        pass

    def _get_schema(self):
        pass


class JSONLinesFileSource(DataSource):
    """
    Read a JSONL (https://jsonlines.org/) file and return a list of objects,
    each being valid json object (e.g. a dictionary or list)
    """

    name = "jsonl"
    version = "0.0.1"
    container = "python"

    def __init__(
        self,
        urlpath: str,
        text_mode: bool = True,
        text_encoding: str = "utf8",
        compression: str = None,
        read: bool = True,
        metadata: dict = None,
        storage_options: dict = None,
    ):
        """
        Parameters
        ----------
        urlpath : str
            Target file. Can include protocol specified (e.g., "s3://").
        text_mode : bool
            Whether to open the file in text mode, recoding binary
            characters on the fly
        text_encoding : str
            If text_mode is True, apply this encoding. UTF* is by far the most
            common
        compression : str or None
            If given, decompress the file with the given codec on load. Can
            be something like "zip", "gzip", "bz2", or to try to guess from the
            filename, 'infer'.
        storage_options: dict
            Options to pass to the file reader backend, including text-specific
            encoding arguments, and parameters specific to the remote
            file-system driver, if using.
        """
        from fsspec.utils import compressions

        VALID_COMPRESSIONS = list(compressions.values()) + ["infer"]

        self._urlpath = urlpath
        self._storage_options = storage_options or {}
        self._dataframe = None
        self._file = None
        self.compression = compression
        if compression is not None:
            if compression not in VALID_COMPRESSIONS:
                raise ValueError(
                    f"Compression value {compression} must be one of {VALID_COMPRESSIONS}"
                )
        self.mode = "rt" if text_mode else "rb"
        self.encoding = text_encoding
        self._read = read
        super().__init__(metadata=metadata)

    @contextlib.contextmanager
    def _open(self):
        """
        Yields an fsspec.OpenFile object
        """
        import fsspec

        urlpath = self._get_cache(self._urlpath)[0]
        with fsspec.open(
            urlpath,
            mode=self.mode,
            encoding=self.encoding,
            compression=self.compression,
            **self._storage_options,
        ) as f:
            yield f

    def read(self):
        with self._open() as f:
            return list(map(json.loads, f))

    def head(self, nrows: int = 100):
        """
        return the first `nrows` lines from the file
        """
        with self._open() as f:
            return list(map(json.loads, islice(f, nrows)))

    def _load_metadata(self):
        pass

    def _get_schema(self):
        pass



print("import: 'intake'")
import intake

print("import: 'intake.auth'")
import intake.auth

print("import: 'intake.catalog'")
import intake.catalog

print("import: 'intake.source'")
import intake.source




#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------


class CatalogException(Exception):
    """Basic exception for errors raised by catalog"""


class PermissionDenied(CatalogException):
    """Raised when user requests functionality that they do not have permission
    to access.
    """


class ShellPermissionDenied(PermissionDenied):
    """The user does not have permission to execute shell commands."""
    def __init__(self, msg=None):
        if msg is None:
            msg = "Additional permissions needed to execute shell commands."
        super(ShellPermissionDenied, self).__init__(msg)


class EnvironmentPermissionDenied(PermissionDenied):
    """The user does not have permission to read environment variables."""
    def __init__(self, msg=None):
        if msg is None:
            msg = "Additional permissions needed to read environment variables."
        super(EnvironmentPermissionDenied, self).__init__(msg)


class ValidationError(CatalogException):
    """Something's wrong with the catalog spec"""
    def __init__(self, message, errors):
        super(ValidationError, self).__init__(message)
        self.errors = errors


class DuplicateKeyError(ValidationError):
    """Catalog contains key duplications"""
    def __init__(self, context, context_mark, problem, problem_mark):
        line = problem_mark.line
        column = problem_mark.column
        msg = "duplicate key found on line {}, column {}".format(
            line + 1, column + 1)
        super(DuplicateKeyError, self).__init__(msg, [])


class ObsoleteError(ValidationError):
    pass


class ObsoleteParameterError(ObsoleteError):
    def __init__(self):
        msg = """Detected old syntax. See details for upgrade instructions to new syntax:

[old syntax]

parameters:
  - name: abc
    type: str

[new syntax]

parameters:
  abc:
    type: str
"""
        super(ObsoleteParameterError, self).__init__(msg, [])


class ObsoleteDataSourceError(ObsoleteError):
    def __init__(self):
        msg = """Detected old syntax. See details for upgrade instructions to new syntax:

[old syntax]

sources:
  - name: abc
    driver: csv

[new syntax]

sources:
  abc:
    driver: csv
"""
        super(ObsoleteDataSourceError, self).__init__(msg, [])



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2019, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import panel as pn
from ..base import Base, MAX_WIDTH


class SearchInputs(Base):
    """Input areas to control search parameters

    Attributes
    ----------
    text: str
        text to use in search. Displayed in text_widget
    depth: int or 'All'
        depth of nestedness to use in search. Displayed in depth_widget.
    children: list of panel objects
        children that will be used to populate the panel when visible
    panel: panel layout object
        instance of a panel layout (row or column) that contains children
        when visible
    watchers: list of param watchers
        watchers that are set on children - cleaned up when visible
        is set to false.
    """
    def __init__(self, **kwargs):
        self.panel = pn.Row(name='Search Inputs', height_policy='min', margin=0)
        super().__init__(**kwargs)

    def setup(self):
        text_label = pn.pane.Markdown('Search Text:', align='center')
        depth_label = pn.pane.Markdown('Depth:', align='center')
        self.text_widget = pn.widgets.TextInput(
            placeholder="Set of words",
            width_policy='max',
            align='center')
        self.depth_widget = pn.widgets.Select(
            options=['1', '2', '3', '4', '5', 'All'],
            width=80, height=30,
            align='center')

        self.children = [text_label, self.text_widget,
                         depth_label, self.depth_widget]

    @property
    def text(self):
        """Text to use in search. Displayed in text_widget"""
        return self.text_widget.value

    @text.setter
    def text(self, text):
        self.text_widget.value = text

    @property
    def depth(self):
        """Depth of nestedness to use in search. Displayed in depth_widget"""
        return int(self.depth_widget.value) if self.depth_widget.value != 'All' else 99

    @depth.setter
    def depth(self, depth):
        if isinstance(depth, int):
            depth = str(depth) if depth <=5 else 'All'
        self.depth_widget.value = depth

    def __getstate__(self):
        """Serialize the current state of the object"""
        return {
            'text': self.text,
            'depth': self.depth,
        }

    def __setstate__(self, state):
        """Set the current state of the object from the serialized version.
        Works inplace. See ``__getstate__`` to get serialized version and
        ``from_state`` to create a new object."""
        self.text = state['text']
        self.depth = state['depth']
        return self


class Search(Base):
    """Search panel for searching a list of catalogs

    Parameters
    ----------
    cats: list of catalogs
        catalogs that will be searched
    done_callback: function with cats as input
        function that is called when new cats have been generated
        via the search functionality

    Attributes
    ----------
    children: list of panel objects
        children that will be used to populate the panel when visible
    panel: panel layout object
        instance of a panel layout (row or column) that contains children
        when visible
    watchers: list of param watchers
        watchers that are set on children - cleaned up when visible
        is set to false.
    """
    def __init__(self, cats, done_callback=None, **kwargs):
        self.cats = cats
        self.done_callback = done_callback
        self.panel = pn.Row(name='Search',
                            height_policy='min',
                            width_policy='max',
                            max_width=MAX_WIDTH,
                            margin=0)
        self.inputs = SearchInputs()
        super().__init__(**kwargs)

    def setup(self):
        self.widget = pn.widgets.Button(name='🔍', width=50, align='center')

        self.watchers = [
            self.widget.param.watch(self.do_search, 'clicks')
        ]

        self.children = [self.inputs.panel, self.widget]

    def do_search(self, arg=None):
        """Do search and close panel"""
        new_cats = []
        for cat in self.cats:
            new_cat = cat.search(self.inputs.text,
                                 depth=self.inputs.depth)
            if len(list(new_cat)) > 0:
                new_cats.append(new_cat)
        if len(new_cats) > 0:
            self.done_callback(new_cats)
            self.visible = False

    def __getstate__(self, include_cats=True):
        """Serialize the current state of the object. Set include_cats
        to False when using with another panel that will include cats."""
        state = {
            'visible': self.visible,
            'inputs': self.inputs.__getstate__()
        }
        if include_cats:
            state['cats'] = [cat.__getstate__() for cat in self.cats]
        return state

    def __setstate__(self, state):
        """Set the current state of the object from the serialized version.
        Works inplace. See ``__getstate__`` to get serialized version and
        ``from_state`` to create a new object."""
        if 'cats' in state:
            self.cats = state['cats']
        self.visible = state.get('visible', True)
        self.inputs.__setstate__(state['inputs'])
        return self



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------
import time
from uuid import uuid4

import itertools
import logging
import msgpack
import tornado.gen
import tornado.ioloop
import tornado.web

from intake.config import conf
from intake.container import serializer
from intake.utils import remake_instance
from intake import __version__
from intake.compat import unpack_kwargs, pack_kwargs
logger = logging.getLogger('intake')


class IntakeServer(object):
    """Main intake-server tornado application"""
    def __init__(self, catalog):
        self._catalog = catalog
        self._cache = SourceCache()
        self._periodic_callbacks = []
        auth = conf.get('auth', 'intake.auth.base.BaseAuth')
        logger.debug('auth: %s' % auth)
        self._auth = remake_instance(auth)

    def get_handlers(self):
        return [
            (r"/v1/info", ServerInfoHandler,
             dict(catalog=self._catalog, cache=self._cache, auth=self._auth)),
            (r"/v1/source", ServerSourceHandler,
             dict(catalog=self._catalog, cache=self._cache, auth=self._auth)),
        ]

    def make_app(self):
        handlers = self.get_handlers()
        return tornado.web.Application(handlers)

    def start_periodic_functions(self, close_idle_after=None,
                                 remove_idle_after=None):
        if len(self._periodic_callbacks) > 0:
            raise Exception('Periodic functions already started '
                            'for this server')

        # Disabling periodic timers with None makes testing easier
        if close_idle_after is not None:
            cache_closer = self._make_cache_closer(close_idle_after)
            self._periodic_callbacks.append(cache_closer)

        if remove_idle_after is not None:
            cache_remover = self._make_cache_remover(remove_idle_after)
            self._periodic_callbacks.append(cache_remover)

        for callback in self._periodic_callbacks:
            callback.start()

    def _make_cache_closer(self, idle_time):
        def cache_closer_callback():
            self._cache.close_idle(idle_time)
        return self._make_cache_callback(cache_closer_callback, idle_time)

    def _make_cache_remover(self, idle_time):
        def cache_remover_callback():
            self._cache.remove_idle(idle_time)
        return self._make_cache_callback(cache_remover_callback, idle_time)

    def _make_cache_callback(self, callback, idle_time):
        # Check ever 1/10 of the idle_time
        interval_ms = (idle_time / 10.0) * 1000
        callback = tornado.ioloop.PeriodicCallback(callback, interval_ms)
        return callback


class ServerInfoHandler(tornado.web.RequestHandler):
    """Basic info about the server"""
    def initialize(self, cache, catalog, auth):
        self.cache = cache
        self.catalog = catalog
        self.auth = auth

    def get(self):
        head = self.request.headers
        page_size = self.get_argument('page_size', None)
        page_offset = self.get_argument('page_offset', 0)
        if self.auth.allow_connect(head):
            if 'source-id' in head:
                cat = self.cache.get(head['source-id'])
            else:
                cat = self.catalog
            sources = []
            if page_size is None:
                # Return all the results in one page. This is important for
                # back-compat with clients that predate pagination. It may
                # also be useful to keep things simple for clients that do not
                # need pagination.
                start = stop = None
            else:
                start = int(page_offset)
                stop = int(page_offset) + int(page_size)
            page = itertools.islice(cat.items(), start, stop)
            for name, source in page:
                if self.auth.allow_access(head, source, self.catalog):
                    info = source.describe().copy()
                    modified_args = info['args'].copy()
                    info['name'] = name
                    for k, v in info['args'].items():
                        try:
                            msgpack.packb(v, **pack_kwargs)
                        except TypeError:
                            modified_args[k] = 'UNSERIALIZABLE_VALUE'
                    info['args'] = modified_args
                    sources.append(info)
            try:
                length = len(cat)
            except TypeError:
                length = sum(1 for entry in cat)
            server_info = dict(version=__version__, sources=sources,
                               length=length,
                               metadata=cat.metadata)
        else:
            msg = 'Access forbidden'
            raise tornado.web.HTTPError(status_code=403, log_message=msg,
                                        reason=msg)
        self.write(msgpack.packb(server_info, **pack_kwargs))


class SourceCache(object):
    """Stores DataSources requested by some user"""
    def __init__(self):
        self._sources = {}

    def add(self, source, source_id=None):
        if source_id is None:
            source_id = str(uuid4())
        now = time.time()
        self._sources[source_id] = dict(source=source, open_time=now,
                                        last_time=now)
        logger.debug('Adding %s to cache, uuid %s' % (source, source_id))
        return source_id

    def get(self, uuid):
        record = self._sources[uuid]
        record['last_time'] = time.time()
        return record['source']

    def peek(self, uuid):
        """Get the source but do not change the last access time"""
        return self._sources[uuid]['source']

    def touch(self, uuid):
        record = self._sources[uuid]
        record['last_time'] = time.time()

    def close_idle(self, idle_secs):
        threshold = time.time() - idle_secs

        for record in self._sources.values():
            if record['last_time'] < threshold:
                record['source'].close()

    def remove_idle(self, idle_secs):
        threshold = time.time() - idle_secs

        # Make a copy of the items so we can mutate the dictionary
        for uuid, record in list(self._sources.items()):
            if record['last_time'] < threshold:
                logger.debug('Removing source %s from cache' % uuid)
                del self._sources[uuid]


class ServerSourceHandler(tornado.web.RequestHandler):
    """Open or stream data source

    The requests "action" field (open|read) specified what the request wants
    to do. Open caches the source and created an ID for it, read uses that
    ID to reference the source and read a partition.
    """
    def initialize(self, catalog, cache, auth):
        self._catalog = catalog
        self._cache = cache
        self.auth = auth

    def get(self):
        """
        Access one source's info.

        This is for direct access to an entry by name for random access, which
        is useful to the client when the whole catalog has not first been
        listed and pulled locally (e.g., in the case of pagination).
        """
        head = self.request.headers
        name = self.get_argument('name')
        if self.auth.allow_connect(head):
            if 'source-id' in head:
                cat = self._cache.get(head['source-id'])
            else:
                cat = self._catalog
            try:
                source = cat[name]
            except KeyError:
                msg = 'No such entry'
                raise tornado.web.HTTPError(status_code=404, log_message=msg,
                                            reason=msg)
            if self.auth.allow_access(head, source, self._catalog):
                info = source.describe().copy()
                info['name'] = name

                source_info = dict(source=info)
                try:
                    out = msgpack.packb(source_info, **pack_kwargs)
                except TypeError:
                    info['direct_access'] = 'forbid'
                    modified_args = source_info['source']['args'].copy()
                    for k, v in source_info['source']['args'].items():
                        try:
                            msgpack.packb(v, **pack_kwargs)
                        except TypeError:
                            modified_args[k] = 'UNSERIALIZABLE_VALUE'
                    source_info['source']['args'] = modified_args
                    out = msgpack.packb(source_info, **pack_kwargs)
                self.write(out)
                return

        msg = 'Access forbidden'
        raise tornado.web.HTTPError(status_code=403, log_message=msg,
                                    reason=msg)

    @tornado.gen.coroutine
    def post(self):
        request = msgpack.unpackb(self.request.body, **unpack_kwargs)
        action = request['action']
        head = self.request.headers
        logger.debug('Source POST: %s' % request)
        
        if action == 'search':
            if 'source-id' in head:
                cat = self._cache.get(head['source-id'])
            else:
                cat = self._catalog
            query = request['query']
            # Construct a cache key from the source_id of the Catalog being
            # searched and the query itself.
            query_source_id = '-'.join((head.get('source-id', 'root'),
                                        str(query)))
            try:
                cat = self._cache.get(query_source_id)
            except KeyError:
                try:
                    args, kwargs = query
                    results_cat = cat.search(*args, **kwargs)
                except Exception as err:
                    logger.exception("Search query %r on Catalog %r failed",
                                     query, cat)
                    raise tornado.web.HTTPError(
                        status_code=400,
                        log_message="Search query failed",
                        reason=str(err).replace("\n", " "))
                self._cache.add(results_cat, source_id=query_source_id)
            response = {'source_id': query_source_id}
            self.write(msgpack.packb(response, **pack_kwargs))
            self.finish()
        elif action == 'open':
            if 'source-id' in head:
                cat = self._cache.get(head['source-id'])
            else:
                cat = self._catalog
            entry_name = request['name']
            try:
                entry = cat[entry_name]
            except KeyError:
                msg = "Catalog has no entry {!r}".format(entry_name)
                raise tornado.web.HTTPError(status_code=404, log_message=msg,
                                            reason=msg)
            if not self.auth.allow_access(head, entry, cat):
                msg = 'Access forbidden'
                raise tornado.web.HTTPError(status_code=403, log_message=msg,
                                            reason=msg)
            user_parameters = request.get('parameters', {})
            client_plugins = request.get('available_plugins', [])

            # Can the client directly access the data themselves?
            open_desc = entry.describe()
            direct_access = open_desc['direct_access']
            plugin_name = open_desc['plugin']
            if isinstance(plugin_name, list):
                for pl in plugin_name:
                    if pl in client_plugins:
                        plugin_name = pl
                        break
            client_has_plugin = plugin_name in client_plugins

            if direct_access == 'forbid' or \
                    (direct_access == 'allow' and not client_has_plugin):
                logger.debug("Opening entry %s" % entry)
                source = entry.configure_new(**user_parameters)
                try:
                    source.on_server = True
                    source.discover()
                except Exception as e:
                    import traceback
                    traceback.print_exc()
                    raise tornado.web.HTTPError(status_code=400,
                                                log_message="Discover failed",
                                                reason=str(e).replace("\n", " "))
                source_id = self._cache.add(source)
                logger.debug('Container: %s, ID: %s' % (source.container,
                                                        source_id))
                response = dict(source._schema or {})
                response.update(dict(container=source.container,
                                     source_id=source_id,
                                     metadata=source.metadata))
                self.write(msgpack.packb(response, **pack_kwargs))
                self.finish()
            elif direct_access == 'force' and not client_has_plugin:
                msg = 'client must have plugin "%s" to access source "%s"' \
                      '' % (plugin_name, entry_name)
                raise tornado.web.HTTPError(status_code=400, log_message=msg,
                                            reason=msg)
            else:
                # If we get here, the client can access the source directly
                # some server-side args need to be parsed
                response = open_desc
                user_parameters['plugin'] = plugin_name
                response['args'] = (entry._entry._create_open_args(user_parameters)[1])
                self.write(msgpack.packb(response, **pack_kwargs))
                self.finish()

        elif action == 'read':
            source_id = request['source_id']
            source = self._cache.get(source_id)
            accepted_formats = request['accepted_formats']
            accepted_compression = request.get('accepted_compression', ['none'])
            partition = request.get('partition', None)

            chunk_encoder = self._pick_encoder(accepted_formats,
                                               accepted_compression,
                                               source.container)

            logger.debug("Read partition %s" % partition)
            if partition is not None:
                chunk = source.read_partition(partition)
            else:
                assert source.npartitions < 2
                chunk = source.read()

            data = chunk_encoder.encode(chunk, source.container)
            msg = dict(format=chunk_encoder.format_name,
                       compression=chunk_encoder.compressor_name,
                       container=source.container, data=data)
            self.write(msgpack.packb(msg, **pack_kwargs))
            self.flush()
            self._cache.touch(source_id)  # keep source alive

            self.finish()

        else:
            msg = '"%s" not a valid source action' % action
            raise tornado.web.HTTPError(status_code=400, log_message=msg,
                                        reason=msg)

    def _pick_encoder(self, accepted_formats, accepted_compression, container):
        format_encoder = None

        for f in accepted_formats:
            if f in serializer.format_registry:
                format_encoder = serializer.format_registry[f]

        if format_encoder is None:
            msg = 'Unable to find compatible format'
            raise tornado.web.HTTPError(status_code=400, log_message=msg,
                                        reason=msg)

        compressor = serializer.NoneCompressor()  # Default
        for f in accepted_compression:
            if f in serializer.compression_registry:
                compressor = serializer.compression_registry[f]

        return serializer.ComboSerializer(format_encoder, compressor)



#-----------------------------------------------------------------------------
# Copyright (c) 2012 - 2018, Anaconda, Inc. and Intake contributors
# All rights reserved.
#
# The full license is in the LICENSE file, distributed with this software.
#-----------------------------------------------------------------------------

import ast
from collections.abc import Iterable
import functools
import itertools
from jinja2 import Environment, meta, Undefined
import os
import re
import shlex
import subprocess
import sys


def flatten(iterable):
    """Flatten an arbitrarily deep list"""
    # likely not used
    iterable = iter(iterable)
    while True:
        try:
            item = next(iterable)
        except StopIteration:
            break

        if isinstance(item, str):
            yield item
            continue

        try:
            data = iter(item)
            iterable = itertools.chain(data, iterable)
        except:
            yield item


def reload_on_change(f):
    @functools.wraps(f)
    def wrapper(self, *args, **kwargs):
        self.reload()
        return f(self, *args, **kwargs)

    return wrapper


def clamp(value, lower=0, upper=sys.maxsize):
    """Clamp float between given range"""
    return max(lower, min(upper, value))


def _j_getenv(x, default=""):
    if isinstance(x, Undefined):
        x = x._undefined_name
    if isinstance(default, Undefined):
        default = default._undefined_name
    return os.getenv(x, default)


def _j_getshell(x):
    if isinstance(x, Undefined):
        x = x._undefined_name
    try:
        return subprocess.check_output(x).decode()
    except (IOError, OSError):
        return ""


def _j_passthrough(x, funcname):
    if isinstance(x, Undefined):
        x = x._undefined_name
    return "{{%s(%s)}}" % (funcname, x)


def _expand(p, context, all_vars, client, getenv, getshell):
    if isinstance(p, dict):
        return {k: _expand(v, context, all_vars, client, getenv, getshell)
                for k, v in p.items()}
    elif isinstance(p, (list, tuple, set)):
        return type(p)(_expand(v, context, all_vars, client, getenv, getshell)
                       for v in p)
    elif isinstance(p, str):
        jinja = Environment()
        if getenv and not client:
            jinja.globals['env'] = _j_getenv
        else:
            jinja.globals['env'] = lambda x: _j_passthrough(x, funcname='env')
        if getenv and client:
            jinja.globals['client_env'] = _j_getenv
        else:
            jinja.globals['client_env'] = lambda x: _j_passthrough(x, funcname='client_env')
        if getshell and not client:
            jinja.globals['shell'] = _j_getshell
        else:
            jinja.globals['shell'] = lambda x: _j_passthrough(x, funcname='shell')
        if getshell and client:
            jinja.globals['client_shell'] = _j_getshell
        else:
            jinja.globals['client_shell'] = lambda x: _j_passthrough(x, funcname='client_shell')
        ast = jinja.parse(p)
        all_vars -= meta.find_undeclared_variables(ast)
        return jinja.from_string(p).render(context)
    else:
        # no expansion
        return p


def expand_templates(pars, context, return_left=False, client=False,
                     getenv=True, getshell=True):
    """
    Render variables in context into the set of parameters with jinja2.

    For variables that are not strings, nothing happens.

    Parameters
    ----------
    pars: dict
        values are strings containing some jinja2 controls
    context: dict
        values to use while rendering
    return_left: bool
        whether to return the set of variables in context that were not used
        in rendering parameters

    Returns
    -------
    dict with the same keys as ``pars``, but updated values; optionally also
    return set of unused parameter names.
    """
    all_vars = set(context)
    out = _expand(pars, context, all_vars, client, getenv, getshell)
    if return_left:
        return out, all_vars
    return out


def expand_defaults(default, client=False, getenv=True, getshell=True):
    """Compile env, client_env, shell and client_shell commands

    Execution rules:
    - env() and shell() execute where the cat is loaded, if getenv and getshell
      are True, respectively
    - client_env() and client_shell() execute only if client is True and
      getenv/getshell are also True.

    If both getenv and getshell are False, this method does nothing.

    If the environment variable is missing or the shell command fails, the
    output is an empty string.
    """
    r = re.match(r'env\((.*),?(.*)\)', default)
    if r and not client and getenv:
        gs = r.groups()
        default = os.environ.get(gs[0], gs[1] if len(gs) > 1 else "")
    r = re.match(r'client_env\((.*)\)', default)
    if r and client and getenv:
        default = os.environ.get(r.groups()[0], '')
    r = re.match(r'shell\((.*)\)', default)
    if r and not client and getshell:
        try:
            cmd = shlex.split(r.groups()[0])
            default = subprocess.check_output(
                cmd).rstrip().decode('utf8')
        except (subprocess.CalledProcessError, OSError):
            default = ''
    r = re.match(r'client_shell\((.*)\)', default)
    if r and client and getshell:
        try:
            cmd = shlex.split(r.groups()[0])
            default = subprocess.check_output(
                cmd).rstrip().decode('utf8')
        except (subprocess.CalledProcessError, OSError):
            default = ''
    return default


def merge_pars(params, user_inputs, spec_pars, client=False, getenv=True,
               getshell=True):
    """Produce open arguments by merging various inputs

    This function is called in the context of a catalog entry, when finalising
    the arguments for instantiating the corresponding data source.

    The three sets of inputs to be considered are:
    - the arguments section of the original spec (params)
    - UserParameters associated with the entry (spec_pars)
    - explicit arguments provided at instantiation time, like entry(arg=value)
      (user_inputs)

    Both spec_pars and user_inputs can be considered as template variables and
    used in expanding string values in params.

    The default value of a spec_par, if given, may have embedded env and shell
    functions, which will be evaluated before use, if the default is used and
    the corresponding getenv/getsgell are set. Similarly, string value params
    will also have access to these functions within jinja template groups,
    as well as full jinja processing.

    Where a key exists in both the spec_pars and the user_inputs, the
    user_input wins. Where user_inputs contains keys not seen elsewhere, they
    are regarded as extra kwargs to pass to the data source.

    Where spec pars have the same name as keys in params, their type, max/min
    and allowed fields are used to validate the final values of the
    corresponding arguments.

    Parameters
    ----------
    params : dict
        From the entry's original spec
    user_inputs : dict
        Provided by the user/calling function
    spec_pars : list of UserParameters
        Default and validation instances
    client : bool
        Whether this is all running on a client to a remote server - sets
        which of the env/shell functions are in operation.
    getenv : bool
        Whether to allow pulling environment variables. If False, the
        template blocks will pass through unevaluated
    getshell : bool
        Whether or not to allow executing of shell commands. If False, the
        template blocks will pass through unevaluated

    Returns
    -------
    Final parameter dict
    """
    context = params.copy()
    for par in spec_pars:
        val = user_inputs.get(par.name, par.default)
        if val is not None:
            if isinstance(val, str):
                val = expand_defaults(val, getenv=getenv, getshell=getshell,
                                      client=client)
            context[par.name] = par.validate(val)
    context.update({k: v for k, v in user_inputs.items() if k not in context})
    out, left = expand_templates(params, context, True, client, getenv,
                                 getshell)
    context = {k: v for k, v in context.items() if k in left}
    for par in spec_pars:
        if par.name in context:
            # coerces to type
            context[par.name] = par.validate(context[par.name])
            left.remove(par.name)

    params.update(out)
    user_inputs = expand_templates(user_inputs, context, False, client, getenv,
                                   getshell)
    params.update({k: v for k, v in user_inputs.items() if k in left})
    params.pop('CATALOG_DIR', None)
    for k, v in params.copy().items():
        # final validation/coersion
        for sp in [p for p in spec_pars if p.name == k]:
            params[k] = sp.validate(params[k])

    return params


def coerce_datetime(v=None):
    import pandas

    if not v:
        v = 0

    try:
        iter(v)
    except TypeError: # not iterable
        pass
    else:
        if "__datetime__" in v:
            v = v["as_str"]

    return pandas.to_datetime(v)


COERCION_RULES = {
    'bool': bool,
    'datetime': coerce_datetime,
    'dict': dict,
    'float': float,
    'tuple': tuple,
    'mlist': list,
    'int': int,
    'list': list,
    'str': str,
    'unicode': str,
    'other': lambda x: x
}


def coerce(dtype, value):
    """
    Convert a value to a specific type.

    If the value is already the given type, then the original value is
    returned. If the value is None, then the default value given by the
    type constructor is returned. Otherwise, the type constructor converts
    and returns the value.
    """
    if "[" in dtype:
        dtype, inner = dtype.split("[")
        inner = inner.rstrip("]")
    else:
        inner = None
    if dtype is None:
        return value
    if type(value).__name__ == dtype:
        return value
    if dtype == "mlist":
        if isinstance(value, (tuple, set, dict)):
            return list(value)
        if isinstance(value, str):
            try:
                value = ast.literal_eval(value)
                return list(value)
            except ValueError as e:
                raise ValueError("Failed to coerce string to list") from e
        return value
    op = COERCION_RULES[dtype]
    out = op() if value is None else op(value)
    if isinstance(out, dict) and inner is not None:
        # TODO: recurse into coerce here, to allow list[list[str]] and such?
        out = {k: COERCION_RULES[inner](v) for k, v in out.items()}
    if isinstance(out, (tuple, list, set)) and inner is not None:
        out = op(COERCION_RULES[inner](v) for v in out)
    return out


class RemoteCatalogError(Exception):
    pass


def _has_catalog_dir(args):
    """Check is any value in args dict needs CATALOG_DIR variable"""
    env = Environment()
    for k, arg in args.items():
        parsed_content = env.parse(arg)
        vars = meta.find_undeclared_variables(parsed_content)
        if 'CATALOG_DIR' in vars:
            return True
    return False
